﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using UnityEngine;
using SimpleJSON; // JSONNode

namespace JustAnotherUser {
    public class TriggerIncrementer : MVRScript {
        private List<DAZMorph> _morphs;
        private string _colliderName; // the object that will colide with the collider region
        private MotionAnimationControl _head;
        private DAZCharacterSelector _characterSelector;

        private JSONStorableBool _activeStorable;
        private JSONStorableStringChooser _colliderStorable;
        private JSONStorableFloat _durationStorable;

        private IDictionary<DAZMorph, JSONStorableFloat> _morphIncrement; // all the morphs that it should modify and its increment

        private float _finalTime = 0f; // play the animation until finalTime
        private Atom _collisionBox; // area where to detect collider

        private static readonly string COLLISION_BOX_NAME = "TI_collision_{plugin_id}";
        private static readonly float COLLISION_BOX_SIZE = 0.4f;

        public override void Init() {
            // plugin VaM GUI description
            pluginLabelJSON.val = "TriggerIncrementer v0.1";

            // get the objects
            if (containingAtom.type != "Person" || (this._characterSelector = containingAtom.GetComponentInChildren<DAZCharacterSelector>()) == null) {
                throw new InvalidOperationException("Missing DAZCharacterSelector");
            }

            if ((this._head = GetHead()) == null) throw new InvalidOperationException("Head not found in added object");

            this._morphs = new List<DAZMorph>();
            this._morphIncrement = new Dictionary<DAZMorph, JSONStorableFloat>();
        }


        // Runs once when plugin loads (after Init)
        protected void Start() {
            this._morphs.Clear();
            ScanBank(this._characterSelector.morphBank1, this._morphs); // @author https://github.com/ProjectCanyon/morph-merger/blob/master/MorphMerger.cs

            this._colliderStorable = new JSONStorableStringChooser("collider", null, "", "Collider", (string colliderName) => {
                this._colliderName = colliderName;

                JSONStorable trigger = this._collisionBox.GetStorableByID("Trigger");
                JSONClass triggerJSON = trigger.GetJSON();

                if (!triggerJSON.HasKey("atomFilter")) triggerJSON.Add("atomFilter", "");
                triggerJSON["atomFilter"].Value = this._colliderName;

                if (triggerJSON["trigger"]["startActions"].AsArray.Count == 0)
                {
                    triggerJSON["trigger"]["startActions"][0].Add("receiverAtom", "");
                    triggerJSON["trigger"]["startActions"][0].Add("receiver", "");
                    triggerJSON["trigger"]["startActions"][0].Add("receiverTargetName", "");
                    triggerJSON["trigger"]["startActions"][0].Add("boolValue", "");
                }
                triggerJSON["trigger"]["startActions"][0]["receiverAtom"].Value = containingAtom.name;
                triggerJSON["trigger"]["startActions"][0]["receiver"].Value = this.storeId; // plugin ID
                triggerJSON["trigger"]["startActions"][0]["receiverTargetName"].Value = "run";
                triggerJSON["trigger"]["startActions"][0]["boolValue"].Value = "true";
                trigger.LateRestoreFromJSON(triggerJSON);

                //SuperController.LogMessage("New collider");
            });
            RegisterStringChooser(this._colliderStorable);
            var linkPopup = CreateFilterablePopup(this._colliderStorable);
            linkPopup.popupPanelHeight = 600f;
            linkPopup.popup.onOpenPopupHandlers += () => { this._colliderStorable.choices = SuperController.singleton.GetAtoms().Select(a => a.name).Distinct().ToList(); };

            JSONStorableStringChooser morphChooseList = new JSONStorableStringChooser("morph", this._morphs.Select(m => m.morphName).Distinct().ToList(), "", "Add morph", (string morphName) => {
                if (this._morphIncrement.Keys.Select(e => e.morphName).Contains(morphName)) {
                    SuperController.LogMessage(morphName + " already added");
                    return;
                }

                SuperController.LogMessage("Added " + morphName);
                AddMorph(morphName);
            });
            // we don't care about this information
            linkPopup = CreateFilterablePopup(morphChooseList);
            linkPopup.popupPanelHeight = 600f;

            this._durationStorable = new JSONStorableFloat("duration", 10.0f, 0.1f, 120.0f);
            RegisterFloat(this._durationStorable);
            CreateSlider(this._durationStorable, true); // create on the right

            this._activeStorable = new JSONStorableBool("run", false, (val) => {
                if (!val) return;

                OnCollision();
                this._activeStorable.val = false;
            });
            RegisterBool(this._activeStorable); // we don't care about this information, but we need to register it to set it on the trigger
            CreateToggle(this._activeStorable, true); // create it on the right

            StartCoroutine(GetCollisionBox(COLLISION_BOX_NAME.Replace("{plugin_id}", this.storeId)));

            //SuperController.LogMessage(pluginLabelJSON.val + " Loaded");

            LoadJson(GetPluginJsonFromSave());
        }

        public void OnCollision() {
            SuperController.singleton.GetAtoms().FirstOrDefault(a => a.name == this._colliderName)?.Remove();
            if (this._finalTime > Time.fixedTime) this._finalTime += this._durationStorable.val;
            else this._finalTime = Time.fixedTime + this._durationStorable.val;
        }

        public void Update() {
            if (this._collisionBox == null) return; // wait to load
            this._collisionBox.mainController.transform.position = this._head.transform.TransformPoint(Vector3.down * 0.05f);

            if (this._finalTime < Time.fixedTime) return;
            float secondsSinceLastUpdate = Time.deltaTime;
            foreach (KeyValuePair<DAZMorph,JSONStorableFloat> entry in this._morphIncrement) {
                float addPerAnimation = entry.Value.val,
                    addPerSecond = addPerAnimation / this._durationStorable.val;
                entry.Key.SetValue(entry.Key.appliedValue + addPerSecond*secondsSinceLastUpdate);
            }
            // TODO check GetFloatJSONParamMaxValue (?)
        }

        public float GetDuration() {
            return this._durationStorable.val;
        }

        private IEnumerator GetCollisionBox(string name) {
            // does it already exists?
            this._collisionBox = SuperController.singleton.GetAtoms().FirstOrDefault(a => a.name == name);
            if (this._collisionBox == null) {
                // no collision box; generate a new one
                SuperController.LogMessage("Generating collision box...");

                yield return SuperController.singleton.AddAtomByType("CollisionTrigger", name);
                this._collisionBox = SuperController.singleton.GetAtoms().FirstOrDefault(a => a.name == name);

                // modify the new collision box
                this._collisionBox.GetStorableByID("scale").GetFloatJSONParam("scale").val = COLLISION_BOX_SIZE;
            }
        }

        private void AddMorph(string morphName, float value = 0.0f) {
            DAZMorph morph = FindMorphByName(this._morphs, morphName);
            if (morph == null) return;

            JSONStorableFloat jsonFloat = new JSONStorableFloat(morphName, value, -1.0f, 1.0f);
            RegisterFloat(jsonFloat);
            CreateSlider(jsonFloat);

            this._morphIncrement.Add(morph, jsonFloat);
        }

        private void LoadJson(JSONNode node) {
            if (node == null) return;

            foreach (string entry in node.AsObject.Keys) {
                switch (entry) {
                    case "collider":
                        this._colliderStorable.val = node[entry].ToString();
                        break;

                    case "duration":
                        this._durationStorable.val = node[entry].AsFloat;
                        break;

                    case "id":
                    case "pluginLabel":
                        break; // ignore

                    default:
                        // tracked morphs
                        try {
                            AddMorph(entry, node[entry].AsFloat);
                        } catch (Exception ex) { }
                        break;
                }
            }
        }

        // @author https://raw.githubusercontent.com/ChrisTopherTa54321/VamScripts/master/FloatMultiParamRandomizer.cs
        public JSONNode GetPluginJsonFromSave() {
            foreach (JSONNode atoms in SuperController.singleton.loadJson["atoms"].AsArray) {
                if (!atoms["id"].Value.Equals(containingAtom.name)) continue;

                foreach (JSONNode storable in atoms["storables"].AsArray) {
                    if (storable["id"].Value == this.storeId) {
                        return storable;
                    }
                }
            }

            return null;
        }

        private MotionAnimationControl GetHead() {
            foreach (MotionAnimationControl mac in containingAtom.motionAnimationControls) { // TODO get head inside linkableRigidbodies?
                if (!mac.name.Equals("headControl")) continue;
                return mac;
            }

            return null; // not found
        }

        private void ScanBank(DAZMorphBank bank, List<DAZMorph> morphs) { // TODO only morph (not morph & pose)
            if (bank == null) return;

            foreach (DAZMorph morph in bank.morphs) {
                if (!morph.visible) continue;

                morphs.Add(morph);
                //SuperController.LogMessage(morph.morphName);
            }
        }

        private DAZMorph FindMorphByName(List<DAZMorph> morphs, string name) {
            foreach (DAZMorph morph in morphs) {
                if (!morph.morphName.Equals(name)) continue;

                return morph;
            }

            return null; // not found
        }
    }
}